import asyncio
import json
import logging

import aiohttp
import requests

from gqla.Executor.abstracts import AbstractExecutor, AbstractRunner


class AsyncRunner(AbstractRunner):

    def __init__(self):
        self.url = None

    def set_url(self, url):
        self.url = url

    def _can_query(self):
        if self.url is None:
            raise AttributeError

    async def run(self, pid, query):
        self._can_query()
        logging.info('Fetch async process {} started'.format(pid))
        async with aiohttp.request('POST', self.url, json=query) as resp:
            response = await resp.text()
        logging.info('Fetch async process {} ended'.format(pid))
        return json.loads(response)


class SyncRunner(AbstractRunner):

    def _can_query(self):
        if self.url is None:
            raise AttributeError

    def __init__(self):
        self.url = None

    def set_url(self, url):
        self.url = url

    def run(self, pid, query):
        self._can_query()
        logging.info('Fetch sync process {} started'.format(pid))
        response = requests.post(self.url, json=query)
        logging.info('Fetch async process {} ended'.format(pid))
        return json.loads(response.text)


class BasicExecutor(AbstractExecutor):

    def __init__(self, url, port, storage, raw="query {{ {query} }}", template="http://{}:{}/graphql", runner=AsyncRunner()):
        self.url = url
        self.port = port
        self._storage = storage
        self.storage = storage.storage if storage is not None else None
        self.QUERY_RAW = raw
        self.URL_TEMPLATE = template
        self.runner = runner
        runner.set_url(self.URL_TEMPLATE.format(self.url, self.port))

    async def execute(self, pid='N/A', query=None, **kwargs):
        if self._storage is not None:
            self.storage = self._storage.storage
        if query is None:
            raise AttributeError
        if len(kwargs) > 0:
            params = "(" + str(kwargs).replace("'", '').replace('{', '').replace('}', '') + ")"
        else:
            params = ''
        if self.storage is not None:
            if query in self.storage:
                instance = self.storage[query]
                instance.args(params)
                query = instance.query
                logging.debug(query)

                query = {
                    'query': self.QUERY_RAW.format(query=query)
                }
        futures = [self.runner.run(pid, query=query)]
        done, pending = await asyncio.wait(futures)
        result = done.pop().result()
        return result







from gqla.GQQuery.Generator.Generator import BasicQueryGenerator, NormalRule, RecursiveRule
from gqla.GQQuery.GQQueriy import BasicQuery
from gqla.GQQuery.abstracts import AbstractQuery
from gqla.GQQStorage.abstracts import AbstractStorage


class BasicStorage(AbstractStorage):

    def __init__(self):
        self._query = {}

    def add(self, name: str, query: AbstractQuery):
        self._query[name] = query

    @property
    def storage(self):
        return self._query

    def create(self, name, item, model, ignore, recursive_depth):
        generator = BasicQueryGenerator(NormalRule(), RecursiveRule(), model, ignore, recursive_depth)
        query = BasicQuery(name, generator, item)
        query.regenerate()
        self.add(name, query)






from gqla.GQQuery.Generator.abstracts import AbstractRule, AbstractGenerator


class NormalRule(AbstractRule):
    def run(self, item):
        return ""


class RecursiveRule(AbstractRule):
    def run(self, item, vault=None, ignore=None, recursive_depth=5, depth=0):
        query = []
        for field in item.fields:
            if item.fields[field].kind == "OBJECT":
                if field in ignore:
                    continue
                depth += 1
                subquery_val = item.fields[field].name
                subquery_val = vault.items[subquery_val]
                subquery_val = self.run(subquery_val, vault, ignore, depth, recursive_depth)
                depth -= 1
                if subquery_val is None:
                    continue
                query.append((str(field) + ' {' + ' '.join(subquery_val) + '}'))
            else:
                if field in ignore:
                    continue
                query.append(field)
                if depth >= recursive_depth:
                    return query
        return query


class BasicQueryGenerator(AbstractGenerator):
    def __init__(self, normal: AbstractRule, recursive: AbstractRule, vault, ignore, recursive_depth):
        super().__init__(normal, recursive)
        self.vault = vault
        self.ignore = ignore
        self.recursive_depth = recursive_depth

    @property
    def normal(self):
        return self._normal

    @property
    def recursive(self):
        return self._recursive

    def generate(self, item):
        if item.kind == 'OBJECT':
            try:
                subquery_val = self.recursive.run(self.vault.items[item.name], self.vault, self.ignore, self.recursive_depth)
            except RecursionError:
                raise
            return ' {' + ' '.join(subquery_val) + '}'
        else:
            return self.normal.run(self.vault.items[item])



import asyncio
import json

import logging
import logging.config
import os.path

from gqla.Executor.Executor import BasicExecutor
from gqla.GQLModel.GQLModel import GQModel
from gqla.GQLStorage.GQLStorage import TypeFactory
from gqla.GQQStorage.GQQStorage import BasicStorage


class GQLA:
    __slots__ = ('url', 'port', 'name', '_ignore', '_model', '_queries', '_subpid', 'usefolder', 'recursive_depth',
                 '_depth', 'qStorage', 'executor')

    INTROSPECTION = {
        'query': '\n    query IntrospectionQuery {\n      __schema {\n        queryType { name }\n        '
                 'mutationType { name }\n        subscriptionType { name }\n        types {\n          ...FullType\n  '
                 '      }\n        directives {\n          name\n          description\n          locations\n         '
                 ' args {\n            ...InputValue\n          }\n        }\n      }\n    }\n\n    fragment FullType '
                 'on __Type {\n      kind\n      name\n      description\n      fields(includeDeprecated: true) {\n   '
                 '     name\n        description\n        args {\n          ...InputValue\n        }\n        type {'
                 '\n          ...TypeRef\n        }\n        isDeprecated\n        deprecationReason\n      }\n      '
                 'inputFields {\n        ...InputValue\n      }\n      interfaces {\n        ...TypeRef\n      }\n    '
                 '  enumValues(includeDeprecated: true) {\n        name\n        description\n        isDeprecated\n  '
                 '      deprecationReason\n      }\n      possibleTypes {\n        ...TypeRef\n      }\n    }\n\n    '
                 'fragment InputValue on __InputValue {\n      name\n      description\n      type { ...TypeRef }\n   '
                 '   defaultValue\n    }\n\n    fragment TypeRef on __Type {\n      kind\n      name\n      ofType {'
                 '\n        kind\n        name\n        ofType {\n          kind\n          name\n          ofType {'
                 '\n            kind\n            name\n            ofType {\n              kind\n              '
                 'name\n              ofType {\n                kind\n                name\n                ofType {'
                 '\n                  kind\n                  name\n                  ofType {\n                    '
                 'kind\n                    name\n                  }\n                }\n              }\n           '
                 ' }\n          }\n        }\n      }\n    }\n  ',
        'variables': {}, 'operationName': None}

    def __init__(self, name, url=None, port=None, ignore=None, usefolder=False, recursive_depth=5):
        self._subpid = 0
        self._model = None
        self._ignore = ignore
        self.name = name
        self.url = url
        self.port = port
        self.usefolder = usefolder
        self.recursive_depth = recursive_depth
        self.qStorage = None
        self.executor = BasicExecutor(url, port, self.qStorage)

        logging.info(' '.join(['CREATED', 'CLASS', str(self.__class__)]))

    def set_ignore(self, ignore_):
        self._ignore = ignore_

    async def query_one(self, query_name, to_file=False, **kwargs):

        result = await self.executor.execute(self._subpid, query_name)

        if self.usefolder:
            if to_file:
                folder = os.path.join('', self.name)
                filename = os.path.join(folder, '_' + query_name + '.json')
                logging.info(' '.join(['WRITING', query_name, 'RESULT TO', filename]))
                if not os.path.exists(folder):
                    os.mkdir(folder)
                with open(filename, 'w') as ofs:
                    ofs.write(json.dumps(result, indent=4))
        return result

    async def introspection(self):
        logging.info(' '.join(['QUERRYING', self.name, 'INTROSPECTION']))
        result = await self.query_one(self.INTROSPECTION)
        queries = result['data']['__schema']['types']

        self.create_data(queries)
        self.generate_queries()

    def create_data(self, data):
        self._model = GQModel()
        for item in data:
            obj = TypeFactory(item)
            if obj is not None:
                self._model.add_item(obj.parse(item))

    def generate_queries(self, specific=False):
        self.qStorage = BasicStorage()
        if 'Query' in self._model.items:
            queries = self._model.items['Query'].fields
        elif 'Queries' in self._model.items:
            queries = self._model.items['Queries'].fields
        else:
            raise NotImplementedError

        for query in queries:
            self.qStorage.create(query, queries[query], self._model, self._ignore, self.recursive_depth)
        self.executor._storage = self.qStorage


async def asynchronous():  # Пример работы
    helper = GQLA('solar', url='localhost', port='8080', usefolder=True)

    ignore = ['pageInfo', 'deprecationReason', 'isDeprecated', 'cursor', 'parent1']

    helper.set_ignore(ignore)

    await helper.introspection()

    for query in helper.qStorage.storage:
        print(helper.qStorage.storage[query].query)
    result = await helper.query_one('allPlanets')
    print(result)


if __name__ == "__main__":
    from gqla.settings import LOGGING_BASE_CONFIG

    logging.getLogger(__name__)
    logging.config.dictConfig(LOGGING_BASE_CONFIG)
    loop_ = asyncio.get_event_loop()
    loop_.run_until_complete(asynchronous())

    loop_.close()
    pass
